"""
This module defines the possible types in the rule engine, and provides 
conversions from and to Python types.
"""

from calendar import timegm
from collections.abc import Iterable
from datetime import datetime
import clips


class RuleEngineType(object):
    """
    Base class for a possible type
    """
    PYTHON_TYPE = None
    PYTHON_TYPE_CHECK = None
    CLIPS_TYPE = None
    DEFAULT = None

    @classmethod
    def _isinstance(cls, x):
        """ Check if argument is of the current type """
        return isinstance(x, cls.PYTHON_TYPE_CHECK)
    
    @classmethod
    def _from_clips_value(cls, x):
        """ Convert from CLIPS value to Python value """
        return cls.DEFAULT if x is None else cls.PYTHON_TYPE(x)
    
    @classmethod
    def _to_clips_value(cls, x):
        """ Convert from Python value to CLIPS value """
        return x


class Integer(RuleEngineType):
    PYTHON_TYPE = int
    PYTHON_TYPE_CHECK = int
    CLIPS_TYPE = 'INTEGER'
    DEFAULT = 0


class Boolean(Integer):
    PYTHON_TYPE = bool
    PYTHON_TYPE_CHECK = bool
    CLIPS_TYPE = 'SYMBOL'
    TRUE = clips.Symbol('TRUE')
    FALSE = clips.Symbol('FALSE')
    DEFAULT = FALSE

    @classmethod
    def _to_clips_value(cls, x):
        if isinstance(x, clips.Symbol):
            return x
        elif x:
            return cls.TRUE
        else:
            return cls.FALSE
    
    @classmethod
    def _from_clips_value(cls, x):
        return str(x) == 'TRUE'


class Number(RuleEngineType):
    PYTHON_TYPE = float
    PYTHON_TYPE_CHECK = (int, float)
    CLIPS_TYPE = 'NUMBER'
    DEFAULT = 0.0


class String(RuleEngineType):
    PYTHON_TYPE = str
    PYTHON_TYPE_CHECK = str
    CLIPS_TYPE = 'STRING'
    DEFAULT = ''


class Unicode(String):
    pass


class Symbol(String):
    CLIPS_TYPE = 'SYMBOL'
    DEFAULT = ''


class DateTime(RuleEngineType):
    """
    Date-time object. 
    There is no date object in CLIPS, we use an integer representation to
    allow for comparisons and arithmetic.
    """
    PYTHON_TYPE = datetime
    PYTHON_TYPE_CHECK = datetime
    CLIPS_TYPE = 'INTEGER'
    DEFAULT = datetime.fromtimestamp(0)
    
    @classmethod
    def _to_clips_value(cls, x):
        return timegm(x.utctimetuple())

    @classmethod
    def _from_clips_value(cls, x):
        return datetime.utcfromtimestamp(x)
    

class Multifield(RuleEngineType):
    """
    Multifield type. Maps to Python list.
    """
    PYTHON_TYPE = list
    PYTHON_TYPE_CHECK = Iterable
    CLIPS_TYPE = 'MULTIFIELD'
    DEFAULT = ()

    @classmethod
    def _from_clips_value(cls, x):
        if isinstance(x, str) or not isinstance(x, Iterable):
            return x,
        else:
            return tuple(x)


class FactIndexType(RuleEngineType): pass


TYPE_MAP = {t.PYTHON_TYPE: t for t in (Integer, Number, String, Boolean, DateTime)}



import clips

from .actions import Update
from .slots import HasSlots, SlotsMeta
from .expr import BaseExpr, ConvertibleToExpr
from .typedefs import FactIndexType
from .utils import LispExpr, Printable, wrap_clips_errors

RULU_INTERNAL_PREFIX = '_rulu_internal'


class FactReference(ConvertibleToExpr):
    def _update_python_param(self, params, value):
        raise NotImplementedError
    
    def _from_python_params(self, params):
        raise NotImplementedError


class FactMeta(SlotsMeta, FactReference):
    def __getitem__(self, key):
        return self._indexed_reference(key)
    

indexed_references = {}


class Fact(HasSlots, metaclass=FactMeta):
    CLIPS_INSTANCE_TYPE = clips.facts.Fact
            
    @classmethod
    @wrap_clips_errors
    def _build(cls, engine):
        if cls is not Fact: 
            super(Fact, cls)._build(engine)
            lisp = LispExpr('deftemplate', cls._name, cls._slots)
            engine.environment.build(str(lisp))
            cls._clips_type = engine.environment.find_template(cls._name)
        
    @classmethod
    def _to_expr(cls):
        if '_fact_expr' not in cls.__dict__:
            cls._fact_expr = FactExpr(cls)
        return cls._fact_expr

    def _create_clips_obj(self, **kwargs):
        return self._clips_type.assert_fact(**kwargs)

    def _copy_clips_obj(self, obj):
        return obj # Old: obj if hasattr(obj, '_Fact__env') else self._environment.Fact(obj)
    
    def _clips_index(self):
        return self._clips_obj.index
    
    def _delete(self):
        self._clips_obj.retract()
        
    def _update(self, **kwargs):
        for key in self._fields:
            kwargs.setdefault(key, self._clips_obj[key])
        self._clips_obj.retract()
        self._clips_obj = self._init_values(**kwargs)
        self._data = self._init_data()

    @classmethod
    def _update_python_param(cls, params, value):
        params[cls._name] = value
        
    @classmethod
    def _from_python_params(cls, params):
        return params[cls._name]

    @classmethod
    def _indexed_reference(cls, index):
        return indexed_references.setdefault((cls, index), IndexedFactReference(cls, index))


class IndexedFactReference(Printable, FactReference):
    def __init__(self, template_cls, key):
        self._template_cls = template_cls
        self._key = key
        self._fields = {}
        self._fact_expr = FactExpr(self)
        for key, field in template_cls._fields.items():
            new_field = type(field)(_type=field.get_type())
            new_field._init(self, key)
            self._fields[key] = new_field
            setattr(self, key, new_field)

    def _to_expr(self):
        return self._fact_expr

    def __str__(self):
        return '{}[{}]'.format(self._template_cls._name, repr(self._key))
    
    def _update_python_param(self, params, value):
        params.setdefault(self._template_cls._name, {})[self._key] = value
    
    def _from_python_params(self, params):
        return params[self._name][self._key]

    @property
    def _name(self):
        return self._template_cls._name
    

class FactExpr(BaseExpr):
    def __init__(self, fact):
        super().__init__(all_fields=[self])
        self.fact = fact
        
    def __str__(self):
        return str(self.fact)
    
    def get_type(self):
        return FactIndexType

    def to_lisp(self):
        return str(self)
